/*
 * hebe.cpp
 *
 * Copyright (C) 2008 %{AUTHOR} <%{EMAIL}>
 */
#include "hebe.h"
#include "hebeview.h"

#include <QtGui/QDropEvent>
#include <QtGui/QPainter>
#include <QtGui/QPrinter>
#include <QDragEnterEvent>
#include <QDropEvent>
#include <qdebug.h>
#include <qlabel.h>
#include <qtimer.h>
#include <QTime>
#include <qclipboard.h>

#include <kconfigdialog.h>
#include <kfiledialog.h>
#include <kstatusbar.h>
#include <kapplication.h>
#include <kurl.h>
#include <kmenubar.h>
#include <kfontdialog.h>
#include <kmessagebox.h>

#include <kaction.h>
#include <kactioncollection.h>
#include <kstandardaction.h>

#include <KDE/KLocale>
#include <ktoolbar.h>

#include <stdlib.h>

using namespace std;

Hebe::Hebe(): KXmlGuiWindow(),
      mpView(new HebeView(this)),
      mpImages(new HebeVmList())

{
	setCaption( i18n( "Hercules Image Manager" ), false );

	// accept dnd
	setAcceptDrops( true );

	// tell the KXmlGuiWindow that this is indeed the main widget
	setCentralWidget( mpView );

	// check the Hercules program is installed
	checkHerc();

	// then, setup our actions
	setupActions();

	// add a status bar
	setupStatusBar();

	// Enable save/restore of window settings, toolbars, etc.
	setupGUI( Default, "/usr/src/dev/kde4/Hebe/src/hebeui.rc" );
	
	//setupGUI();
	
	// global completion object
	mpCompl = new KCompletion();

	// read saved options
	mConfig = KGlobal::config();
	mpGenGroup = new KConfigGroup( mConfig, "General" );
	readOptions();
	
	// initialise colours
	mRedB.setColor( Qt::red );
	mGreenB.setColor( Qt::green );
	mGrayB.setColor( Qt::lightGray );
	mYellowB.setColor( Qt::yellow );
	mWhiteB.setColor( Qt::white );
	mLightBlueB.setColor( QColor("#74BCEC") );
	
	if ( mpImages->count() > 0 )
		mpView->setCurrentPage( mpImages->VM( 0 )->mpPage );
	
	connect( mpView, SIGNAL(currentPageChanged(KPageWidgetItem*,KPageWidgetItem*)),
					 this, SLOT(pageChanged(KPageWidgetItem*,KPageWidgetItem*)) );
}

Hebe::~Hebe()
{
	shutDown();
	saveOptions();
	delete mpImages;
	delete mpView;
}

bool Hebe::queryClose()
{
	if ( mpImages->activeImages() > 0 )
	{	
		KMessageBox::sorry( this, i18n("There are images still active, shut them down first!") );
		return false;
	}
	else
		return KMainWindow::queryClose();
}

void Hebe::setupActions()
{
	KStandardAction::quit(kapp, SLOT(closeAllWindows()), actionCollection());

	//KStandardAction::preferences(this, SLOT(confOptions()), actionCollection());

	mpAdd = new KAction( KIcon("list-add"), i18n("A&dd Image"), this );
	actionCollection()->addAction( QLatin1String("add"), mpAdd );
	mpAdd->setShortcut( Qt::CTRL+Qt::Key_D );
	mpAdd->setToolTip( i18n("Add Hercules Image") );
	connect( mpAdd, SIGNAL(triggered(bool)), this, SLOT(fileAdd()) );

	mpRemove = new KAction( KIcon("list-remove"), i18n("Remov&e Image"), this );
	actionCollection()->addAction( QLatin1String("remove"), mpRemove );
	mpRemove->setShortcut( Qt::CTRL+Qt::Key_E );
	mpRemove->setToolTip( i18n("Remove Hercules Image") );
	connect( mpRemove, SIGNAL(triggered(bool)), this, SLOT(fileRemove()) );

	mpStart = new KAction( KIcon("system-reboot"), i18n("&Startup"), this );
	actionCollection()->addAction( QLatin1String("startup"), mpStart );
	mpStart->setShortcut( Qt::CTRL+Qt::Key_S );
	mpStart->setToolTip( i18n("Start Hercules image") );
	connect( mpStart, SIGNAL(triggered(bool)), this, SLOT(startSlot()) );

	mpSuspend = new KAction( KIcon("system-suspend"), i18n("S&uspend"), this );
	actionCollection()->addAction( QLatin1String("suspend"), mpSuspend );
	mpSuspend->setShortcut( Qt::CTRL+Qt::Key_U );
	mpSuspend->setToolTip( i18n("Suspend and save image") );
	connect( mpSuspend, SIGNAL(triggered(bool)), this, SLOT(imageCmdSlot()) );

	mpStop = new KAction( KIcon("system-shutdown"), i18n("Shut&down"), this );
	actionCollection()->addAction( QLatin1String("shutdown"), mpStop );
	mpStop->setShortcut( Qt::CTRL+Qt::Key_P );
	mpStop->setToolTip( i18n("Shutdown Hercules image") );
	connect( mpStop, SIGNAL(triggered(bool)), this, SLOT(imageCmdSlot()) );

	mpIPL = new KAction( KIcon("system-run"), i18n("&IPL"), this );
	actionCollection()->addAction( QLatin1String("ipl"), mpIPL );
	mpIPL->setShortcut( Qt::CTRL+Qt::Key_I );
	mpIPL->setToolTip( i18n("IPL selected device") );
	connect( mpIPL, SIGNAL(triggered(bool)), this, SLOT(iplSlot()) );

	mpStartAll = new KAction( KIcon("user-online"), i18n("S&tartAll"), this );
	actionCollection()->addAction( QLatin1String("startall"), mpStartAll );
	mpStartAll->setShortcut( Qt::CTRL+Qt::Key_T );
	mpStartAll->setToolTip( i18n("Start all CPUs") );
	connect( mpStartAll, SIGNAL(triggered(bool)), this, SLOT(imageCmdSlot()) );

	mpStopAll = new KAction( KIcon("user-busy"), i18n("Stop&All"), this );
	actionCollection()->addAction( QLatin1String("stopall"), mpStopAll );
	mpStopAll->setShortcut( Qt::CTRL+Qt::Key_A );
	mpStopAll->setToolTip( i18n("Stop all CPUs") );
	connect( mpStopAll, SIGNAL(triggered(bool)), this, SLOT(imageCmdSlot()) );

	mpRstInt = new KAction( KIcon("task-recurring"), i18n("&Restart"), this );
	actionCollection()->addAction( QLatin1String("rstint"), mpRstInt );
	mpRstInt->setShortcut( Qt::CTRL+Qt::Key_R );
	mpRstInt->setToolTip( i18n("Issue restart interrupt") );
	connect( mpRstInt, SIGNAL(triggered(bool)), this, SLOT(imageCmdSlot()) );

	mpExtInt = new KAction( KIcon("task-attention"), i18n("&ExtInt"), this );
	actionCollection()->addAction( QLatin1String("extint"), mpExtInt );
	mpExtInt->setShortcut( Qt::CTRL+Qt::Key_E );
	mpExtInt->setToolTip( i18n("Issue external interrupt") );
	connect( mpExtInt, SIGNAL(triggered(bool)), this, SLOT(imageCmdSlot()) );

	mpViewIcon = new KAction( KIcon("view-list-icons"), i18n("&Icon"), this );
	actionCollection()->addAction( QLatin1String("icon"), mpViewIcon );
	mpViewIcon->setShortcut( Qt::CTRL+Qt::Key_1 );
	mpViewIcon->setToolTip( i18n("Select device icon view") );
	connect( mpViewIcon, SIGNAL(triggered(bool)), this, SLOT(viewIconSlot(bool)) );

	mpViewTree = new KAction( KIcon("view-list-tree"), i18n("&Tree"), this );
	actionCollection()->addAction( QLatin1String("tree"), mpViewTree );
	mpViewTree->setShortcut( Qt::CTRL+Qt::Key_2 );
	mpViewTree->setToolTip( i18n("Select device tree view") );
	connect( mpViewTree, SIGNAL(triggered(bool)), this, SLOT(viewTreeSlot(bool)) );
	
	mpViewGrp = new QActionGroup( this );
	mpViewGrp->addAction( mpViewIcon );
	mpViewGrp->addAction( mpViewTree );
	mpViewIcon->setCheckable( true );
	mpViewTree->setCheckable( true );
	mpViewTree->setChecked( true );
	mViewMode = TREE;

	mpPageViewList = new KAction( KIcon("view-choose"), i18n("&List"), this );
	actionCollection()->addAction( QLatin1String("page_list"), mpPageViewList );
	mpPageViewList->setShortcut( Qt::CTRL+Qt::Key_3 );
	mpPageViewList->setToolTip( i18n("Select list page view") );
	connect( mpPageViewList, SIGNAL(triggered(bool)), this, SLOT(pageViewSlot(bool)) );
	
	mpPageViewTree = new KAction( KIcon("view-process-all-tree"), i18n("T&ree"), this );
	actionCollection()->addAction( QLatin1String("page_tree"), mpPageViewTree );
	mpPageViewTree->setShortcut( Qt::CTRL+Qt::Key_4 );
	mpPageViewTree->setToolTip( i18n("Select tree page view") );
	connect( mpPageViewTree, SIGNAL(triggered(bool)), this, SLOT(pageViewSlot(bool)) );
	
	mpPageViewTabbed = new KAction( KIcon("view-process-all"), i18n("&Tabbed"), this );
	actionCollection()->addAction( QLatin1String("page_tabbed"), mpPageViewTabbed );
	mpPageViewTabbed->setShortcut( Qt::CTRL+Qt::Key_5 );
	mpPageViewTabbed->setToolTip( i18n("Select tabbed page view") );
	connect( mpPageViewTabbed, SIGNAL(triggered(bool)), this, SLOT(pageViewSlot(bool)) );
	
	mpPageGrp = new QActionGroup( this );
	mpPageGrp->addAction( mpPageViewList );
	mpPageGrp->addAction( mpPageViewTabbed );
	mpPageGrp->addAction( mpPageViewTree );
	mpPageViewList->setCheckable( true );
	mpPageViewTabbed->setCheckable( true );
	mpPageViewTree->setCheckable( true );
	mpPageViewList->setChecked( true );
	
	// initially no VM commands available
	mpStart->setEnabled( false );
	mpStop->setEnabled( false );
	mpSuspend->setEnabled( false );
	mpIPL->setEnabled( false );
	mpStartAll->setEnabled( false );
	mpStopAll->setEnabled( false );
	mpRstInt->setEnabled( false );
	mpExtInt->setEnabled( false );
	
	// show image toolbars
	toolBar( ITB )->show();
	
	// context menu for log entries
	QAction *action;
	mpPopup = new QMenu( mpView );
  action = mpPopup->addAction( KIcon("edit-select-all"), QString(i18n("Select &All")) );
	action->setData( "selectall" );
	action->setShortcut( Qt::CTRL+Qt::Key_A );	
  action = mpPopup->addAction( KIcon("edit-copy"), QString(i18n("&Copy")) );
	action->setData( "copy" );
	action->setShortcut( Qt::CTRL+Qt::Key_C );	
}

void Hebe::setupStatusBar()
{
	mpSBtext1 = new QLabel( i18n("Ready"), this );
	mpSBtext1->setFrameShape( QFrame::StyledPanel );
	mpSBtext1->setFrameShadow( QFrame::Sunken );
	mpSBtext1->setMinimumWidth( 100 );
	statusBar()->addWidget( mpSBtext1 );             	// MIPS
	
	mpSBtext2 = new QLabel( i18n(""), this );
	mpSBtext2->setFrameShape( QFrame::StyledPanel );
	mpSBtext2->setFrameShadow( QFrame::Sunken );
	mpSBtext2->setMinimumWidth( 100 );
	statusBar()->addWidget( mpSBtext2 );						 	// SIOS
	
/*	mpSBtext3 = new QLabel( i18n(""), this );
	mpSBtext3->setFrameShape( QFrame::StyledPanel );
	mpSBtext3->setFrameShadow( QFrame::Sunken );
	mpSBtext3->setMinimumWidth( 70 );
	statusBar()->addWidget( mpSBtext3 );							// SYS*/

	mpSBcb = new QComboBox( this );
	mpSBcb->addItem( "CP00" );
	statusBar()->addWidget( mpSBcb );									// CPU0
	
	mpSBtext4 = new QLabel( i18n(""), this );
	mpSBtext4->setFrameShape( QFrame::StyledPanel );
	mpSBtext4->setFrameShadow( QFrame::Sunken );
	//mpSBtext4->setMinimumWidth( 550 );
	statusBar()->addWidget( mpSBtext4 );							// CPU STATUS
	
	statusBar()->setStyleSheet( "QLabel { font-family: monospace; }" );
	statusBar()->show();
	
	connect( mpSBcb, SIGNAL(currentIndexChanged(int)), this, SLOT(cbIndexChanged(int)) );
}

void Hebe::clearStatusBar()
{
	mpSBtext1->clear();
	mpSBtext2->clear();
	//mpSBtext3->clear();
	mpSBtext4->clear();
}

void Hebe::fileAdd()
{
	// get a new Herc config file
	QString cnf = KFileDialog::getOpenFileName( KUrl(), "*.cnf", this,
							 i18n("Select Hercules configuration file") );
	if ( !cnf.isEmpty() )
		addImage( cnf );
}

void Hebe::fileRemove()
{
	closePage();
}


void Hebe::saveOptions()
{
	mpGenGroup->writeEntry( "files", mpImages->VMnames() );
	mpGenGroup->writeEntry( "comp-items", mpCompl->items() );
	
	if ( mpImages->count() > 0 )
	{	
		const int k = mpView->currentIndex();
		HebeVM *vm = mpImages->VM( k );
		mpGenGroup->writeEntry( "col-sizes", vm->mpCons->mpDevTw->header()->saveState() );
		mpGenGroup->writeEntry( "sizes", vm->mpCons->mpSplitter->sizes() );
	}	
	
	mpGenGroup->writeEntry( "face-type", static_cast<int>( mpView->faceType() ) );
	mpGenGroup->sync();
}

void Hebe::readOptions()
{
	QStringList files, items; 
	int ft;
	
	mPaneSize = mpGenGroup->readEntry( "sizes", QList<int>() );
	if ( mPaneSize[0] > 0 )
	{ 
		mViewMode = ICON;                   // last use was ICON mode
		mpViewIcon->setChecked( true );
		mpViewTree->setChecked( false );
	}	
	
	mColSizes = mpGenGroup->readEntry( "col-sizes", QByteArray() );

  items = mpGenGroup->readEntry( "comp-items", QString( HERC_CMDS ).split( ',' ) );
	if ( !items.isEmpty() )
		mpCompl->setItems( items );
	
	ft = mpGenGroup->readEntry( "face-type", 0 );
	mpView->setFaceType( static_cast<KPageView::FaceType>( ft ) );
	if ( mpView->faceType() != KPageView::Auto )
	{
		mpPageViewList->setChecked( false );
		if ( mpView->faceType() == KPageView::Tabbed )
			mpPageViewTabbed->setChecked( true );
		else if ( mpView->faceType() == KPageView::Tree )
			mpPageViewTree->setChecked( true );
	}	
	
  files = mpGenGroup->readEntry( "files", QStringList() );
	if ( !files.isEmpty() )
	{	
		files.sort();
		for ( int k = 0; k < files.count(); ++k )
			addImage( files[k] );
	}	
}

void Hebe::confOptions()
{
/*	KmConf opts( this );
	if ( opts.exec() == KDialog::Accepted )
	{
		saveOptions();
		fileReload();
	}	*/
}


void Hebe::addImage( const QString &rVMcnf )
{
	HebeVM vm;

	vm.mpCons = new HebeCons( this );
	if ( !mPaneSize.isEmpty() )
		vm.mpCons->mpSplitter->setSizes( mPaneSize );
	if ( !mColSizes.isEmpty() )
		vm.mpCons->mpDevTw->header()->restoreState( mColSizes );
	
	vm.mpCons->mpConsInput->setCompletionObject( mpCompl );
	vm.mpCons->mpDevTw->setSortingEnabled( true );
	vm.mpCons->mpDevIv->setSortingEnabled( true );

  connect( vm.mpCons->mpHercTw, SIGNAL(customContextMenuRequested(const QPoint &)),
     	     this, SLOT(rmbItem(const QPoint&)) );
  connect( vm.mpCons->mpSysTw, SIGNAL(customContextMenuRequested(const QPoint &)),
     	     this, SLOT(rmbItem(const QPoint&)) );
	//connect( vm.mpCons->mpTabs, SIGNAL(currentChanged(int)), this, SLOT(changedTab(int)) );
	connect( vm.mpCons->mpConsInput, SIGNAL(returnPressed(QString)), this, SLOT(cmdInput(QString)) );
	connect( vm.mpCons->mpConsInput, SIGNAL(returnPressed(const QString&)), 
					 mpCompl, SLOT(addItem(const QString&)) );
	connect( vm.mpCons->mpDevTw, SIGNAL(itemDoubleClicked(QTreeWidgetItem*,int)), 
					 this, SLOT(iplDevice(QTreeWidgetItem*,int)) );
	connect( vm.mpCons->mpDevIv, SIGNAL(itemDoubleClicked(QListWidgetItem*)), 
					 this, SLOT(iplDevice(QListWidgetItem*)) );
	
	int p = rVMcnf.lastIndexOf( '/' );
	int q = rVMcnf.lastIndexOf( '/', p-1 );
	if ( q == -1 )
		vm.mName = "RootVM";
	else	
		vm.mName = rVMcnf.mid( q+1, p-q-1 );
	vm.mCnfFile = rVMcnf;
	vm.mDir = rVMcnf.left( p );
	vm.mNumCPUs = getNumCPUs( rVMcnf );
	
	vm.mpPage = new KPageWidgetItem( vm.mpCons, vm.mName );
	vm.mpPage->setHeader( TITLE + vm.mName );
	vm.mpPage->setIcon( KIcon("z10") );
	mpImages->addVM( vm );
	
	mpView->addPage( vm.mpPage );
	mpView->setCurrentPage( vm.mpPage ); 
	mpStart->setEnabled( true );
	mpStop->setEnabled( false );
	mpSuspend->setEnabled( false );
	mpIPL->setEnabled( false );
	mpStartAll->setEnabled( false );
	mpStopAll->setEnabled( false );
	mpRstInt->setEnabled( false );
	mpExtInt->setEnabled( false );
		
	loadSBcb( &vm );
	setViewMode( &vm );
}

void Hebe::startSlot()
{
	const int k = mpView->currentIndex();
	HebeVM *vm = mpImages->VM( k );
	if ( vm == 0 ) return;
	
	if ( !vm->mStarted )
		startProc( vm );
}

void Hebe::imageCmdSlot()
{
	const int k = mpView->currentIndex();
	HebeVM *vm = mpImages->VM( k );
	if ( vm == 0 ) return;
	
	if ( vm->mStarted )
	{	
		if ( QObject::sender() == mpSuspend )
			cmdInput( "suspend", true );
		else if ( QObject::sender() == mpStop )
			cmdInput( "exit", true );
		else if ( QObject::sender() == mpStartAll )
			cmdInput( "startall", true );
		else if ( QObject::sender() == mpStopAll )
			cmdInput( "stopall", true );
		else if ( QObject::sender() == mpRstInt )
			cmdInput( "restart", true );
		else if ( QObject::sender() == mpExtInt )
			cmdInput( "ext", true );
	}	
}

void Hebe::iplSlot()
{
	const int k = mpView->currentIndex();
	HebeVM *vm = mpImages->VM( k );
	if ( vm == 0 ) return;
	
	QString device;
	if ( vm->mStarted )
	{
		if ( mViewMode == ICON )
		{
			QList<QListWidgetItem*> items = vm->mpCons->mpDevIv->selectedItems();
			if ( items.isEmpty() ) return;
			device = items[0]->text();
		}	
		else
		{
			QList<QTreeWidgetItem*> items = vm->mpCons->mpDevTw->selectedItems();
			if ( items.isEmpty() ) return;
			device = items[0]->text( 0 );
		}	
		
		cmdInput( "IPL " + device, true );
	}
}

void Hebe::startProc( HebeVM *vm )
{
	vm->mpProc = new KProcess( this );	
	vm->mpProc->setOutputChannelMode( KProcess::SeparateChannels );
	vm->mpProc->setWorkingDirectory( vm->mDir );

	connect( vm->mpProc, SIGNAL(readyReadStandardOutput()), this, SLOT(readStdOut()) );
	connect( vm->mpProc, SIGNAL(readyReadStandardError()), this, SLOT(readStdErr()) );
	connect( vm->mpProc, SIGNAL(finished(int, QProcess::ExitStatus)), this, 
					 SLOT(finishedProc(int, QProcess::ExitStatus)) );

	*vm->mpProc << mProg << "-f" << vm->mCnfFile << "EXTERNALGUI";
	vm->mpProc->start();
	
	if ( !vm->mpProc->waitForStarted() )
	{
		qDebug() << "error: " << mProg.toAscii() << " " << vm->mName.toAscii() << "failed to start" << "\n";
	}
	else
	{
		mpStart->setEnabled( false );
		mpStop->setEnabled( true );
		mpSuspend->setEnabled( true );
		mpIPL->setEnabled( true );
		mpStartAll->setEnabled( true );
		mpStopAll->setEnabled( true );
		mpRstInt->setEnabled( true );
		mpExtInt->setEnabled( true );
		
		vm->mStarted = true;
		vm->mpCons->mpDevTw->clear();
		vm->mpCons->mpDevIv->clear();
		vm->mpTimer = new QTimer( this );
		vm->mpTimer->start( 2000 );
		connect( vm->mpTimer, SIGNAL(timeout()), this, SLOT(hlTimeout()) );
		
		mpImages->addObjMap( vm->mpProc, vm );
		mpImages->addObjMap( vm->mpTimer, vm );
		
		setViewMode( vm );
		
		checkResumeFile( vm );
	}	
}

void Hebe::finishedProc( int, QProcess::ExitStatus )
{
	HebeVM *vm = mpImages->objMap( QObject::sender() );

	vm->mpProc->close();
	mpImages->removeObjMap( vm->mpProc );
	delete vm->mpProc;
	vm->mpProc = 0;
	vm->mStarted = false;
	vm->mpTimer->stop();
	mpImages->removeObjMap( vm->mpTimer );
	delete vm->mpTimer;

	mpStart->setEnabled( true );
	mpStop->setEnabled( false );
	mpSuspend->setEnabled( false );
	mpIPL->setEnabled( false );
	mpStartAll->setEnabled( false );
	mpStopAll->setEnabled( false );
	mpRstInt->setEnabled( false );
	mpExtInt->setEnabled( false );
	
	if ( vm->mpPage == mpView->currentPage() )
		clearStatusBar();
}

void Hebe::shutDown()
{
	const int n = mpImages->activeImages();
	for ( int k = 0; k < n; ++k )
	{	
		HebeVM *vm = mpImages->VM( k );
		
		if ( vm->mStarted )
		{
			qDebug() << "Shutting down" << vm->mName << flush;
			mpView->setCurrentPage( vm->mpPage );
			cmdInput( "stopall", true );
			cmdInput( "exit", true );
			vm->mpProc->waitForFinished();
		}	
	}	
}

void Hebe::readStdOut()
{
  char buf[4096];  
	HebeVM *vm = mpImages->objMap( QObject::sender() );
	
	vm->mpProc->setReadChannel( QProcess::StandardOutput );
	while ( true )
	{	
		qint64 l = vm->mpProc->readLine( buf, sizeof(buf) );
		if ( l < 1 ) break; 
		
		QString line = QString( buf );
		line.truncate( l-1 );
		
		QTreeWidgetItem *pItem;
		QString msgid = line.left( 9 );
		
		if ( (msgid == "HHC00001I") & !vm->mScpMsgPrfx )
			vm->mScpMsgPrfx = true;
		
		if ( (msgid == "HHC00001I") | ((msgid.left(3) != "HHC") & !vm->mScpMsgPrfx & vm->mIPLed) )
		{
			if ( msgid == "HHC00001I" )
				line.remove( 0, 10 );
			
			pItem = new QTreeWidgetItem( vm->mpCons->mpSysTw );
			pItem->setText( 0, line );
			pItem->setForeground( 0, mGreenB );
			
			splitToolTip( line );	
			pItem->setToolTip( 0, line );
		}	
		else
		{	
			pItem = new QTreeWidgetItem( vm->mpCons->mpHercTw );
		
			if  ( msgid.left(3) == "HHC" )
			{	
				pItem->setText( 0, QTime::currentTime().toString( "hh:mm:ss" ) );
				pItem->setText( 1, line );
				pItem->setForeground( 0, mLightBlueB );
				
				if ( msgid[8] == 'I' ) 			pItem->setForeground( 1, mGreenB );
				else if ( msgid[8] == 'W' ) pItem->setForeground( 1, mYellowB );	
				else if ( msgid[8] == 'A' ) pItem->setForeground( 1, mYellowB );	
				else if ( msgid[8] == 'E' ) pItem->setForeground( 1, mRedB );	
				else if ( msgid[8] == 'S' ) pItem->setForeground( 1, mRedB );	
			}
			else
			{	
				pItem->setText( 0, "  " );
				pItem->setText( 1, line );
				pItem->setForeground( 1, mGrayB );
			}	
		
			splitToolTip( line );	
			pItem->setToolTip( 1, line );
		}	
	}
	
	while ( vm->mpCons->mpHercTw->topLevelItemCount() > MAXLINES )
		delete vm->mpCons->mpHercTw->takeTopLevelItem( 0 );
	while ( vm->mpCons->mpSysTw->topLevelItemCount() > MAXLINES )
		delete vm->mpCons->mpSysTw->takeTopLevelItem( 0 );
	
	vm->mpCons->mpHercTw->resizeColumnToContents( 0 );
	vm->mpCons->mpHercTw->resizeColumnToContents( 1 );
	vm->mpCons->mpSysTw->resizeColumnToContents( 0 );

	vm->mpCons->mpHercTw->scrollToBottom();	
	vm->mpCons->mpSysTw->scrollToBottom();	
}

void Hebe::splitToolTip( QString &line )
{
	for ( int i = 80; i < line.length(); i += 80 )
	{
		int p = line.indexOf( ' ', i );
		if ( p > -1 ) 
			line.replace( p, 1, '\n' );
		else 
		{
			p = line.indexOf( ',', i );
			if ( p > -1 )
				line.insert( p, '\n');
			else
				line.insert( i, '\n' );
		}	
	}	
}

void Hebe::readStdErr()
{
  char buf[4096];
	HebeVM *vm = mpImages->objMap( QObject::sender() );
	vm->mpProc->setReadChannel( QProcess::StandardError );

	while ( true )
	{	
		qint64 l = vm->mpProc->readLine( buf, sizeof(buf) );
		if ( l < 1 ) break;																// no data	

		QString line( QString(buf).left( l-1 ) );
		//qDebug() << line;
		
		if ( vm->mpPage == mpView->currentPage() )
		{	
			if ( line.startsWith( "MIPS=" ) )
				mpSBtext1->setText( line );
			else if ( line.startsWith( "SIOS=" ) )
				mpSBtext2->setText( line );
/*			else if ( line.startsWith( "SYS=" ) )
				mpSBtext3->setText( line );*/
			else if ( line.startsWith( "STATUS=" ) )
				mpSBtext4->setText( line.remove( 0, 7 ) );
		}
		
		if ( line.startsWith( "DEVA=" ) )
			deviceAttach( *vm, line.remove( 0, 5 ) );
		else if ( line.startsWith( "DEVC=" ) )
			deviceChange( *vm, line.mid( 5, 4 ) );
		else if ( line.startsWith( "DEVD=" ) )
			deviceDetach( *vm, line.mid( 5, 4 ) );
		else if ( line.startsWith( "DEVX=" ) )
		{	
			vm->mpCons->mpDevTw->sortItems( 0, Qt::AscendingOrder );
			vm->mpCons->mpDevIv->sortItems( Qt::AscendingOrder );
		}	
	}
}

void Hebe::deviceAttach( HebeVM &vm, const QString &def )
{
	QString model, cua, device;
	QTreeWidgetItem *pItem, *pSubItem;
	QTreeWidget *pTw = vm.mpCons->mpDevTw;
	
	model = def.section( ' ', 1, 2 ).trimmed();
	cua =  def.left( 4 ).trimmed();
	device = def.section( ' ', 4 ).trimmed();
	
	QList<QTreeWidgetItem*> devtype = pTw->findItems( model, Qt::MatchExactly );
	if ( devtype.isEmpty() )
	{
		pItem = new QTreeWidgetItem( pTw );
		pItem->setExpanded( true );
		pItem->setText( 0, model );
		pItem->setForeground( 0, mWhiteB );
		pItem->setIcon( 0, getDevIcon( model ) );
	}	
	else
	{		
		pItem = devtype[0];
	}	

	pSubItem = new QTreeWidgetItem( pItem );
	pSubItem->setText( 0, cua );
	pSubItem->setForeground( 0, mWhiteB );
	pSubItem->setToolTip( 0, i18n("Double-click to IPL from device") );
	pSubItem->setText( 1, device );
	pSubItem->setForeground( 1, mWhiteB );
	pSubItem->setToolTip( 1, device );
	
	QListWidgetItem *pIvItem;
	QListWidget *pLw = vm.mpCons->mpDevIv;
	pIvItem = new QListWidgetItem( cua, pLw );
	pIvItem->setForeground( mWhiteB );
	pIvItem->setToolTip( model + '\n' + device );
	pIvItem->setIcon( getDevIcon( model ) );
}

void Hebe::deviceChange( HebeVM &vm, const QString &device )
{
	if ( mViewMode == TREE )
	{	
		QTreeWidget *pTw = vm.mpCons->mpDevTw;
		QList<QTreeWidgetItem*> pItem = pTw->findItems( device, Qt::MatchFixedString | Qt::MatchRecursive );

		if ( !pItem.isEmpty() )
			pItem[0]->setForeground( 0, mYellowB );
	}
	else
	{
		QListWidget *pLw = vm.mpCons->mpDevIv;
		QList<QListWidgetItem*> pItem = pLw->findItems( device, Qt::MatchFixedString );

		if ( !pItem.isEmpty() )
			pItem[0]->setSelected( true );
	}
}

void Hebe::deviceDetach( HebeVM &vm, const QString &device )
{
	QTreeWidget *pTw = vm.mpCons->mpDevTw;
	QList<QTreeWidgetItem*> pItem = pTw->findItems( device, Qt::MatchFixedString | Qt::MatchRecursive );

	if ( !pItem.isEmpty() )
		delete pItem[0];

	QListWidget *pLw = vm.mpCons->mpDevIv;
	QList<QListWidgetItem*> pIcon = pLw->findItems( device, Qt::MatchFixedString );

	if ( !pIcon.isEmpty() )
		delete pIcon[0];
}


void Hebe::deviceDim( HebeVM &vm )
{
	QTreeWidgetItemIterator it( vm.mpCons->mpDevTw, QTreeWidgetItemIterator::NoChildren );
	while ( *it ) 
	{
		(*it)->setForeground( 0, mWhiteB );
		++it;
	}
	
	vm.mpCons->mpDevIv->clearSelection();
}

KIcon Hebe::getDevIcon( const QString &text )
{
	if ( text.contains( "DASD" ) )
		return KIcon("dasd");
	else if ( text.contains( "PCH" ) )
		return KIcon("punchcard");
	else if ( text.contains( "RDR" ) )
		return KIcon("punchcard");
	else if ( text.contains( "TAPE" ) )
		return KIcon("media-tape");
	else if ( text.contains( "CTC" ) )
		return KIcon("ctc");
	else if ( text.contains( "DSP" ) )
		return KIcon("utilities-terminal");
	else if ( text.contains( "CON" ) )
		return KIcon("terminal");
	else if ( text.contains( "PRT" ) )
		return KIcon("printer");
	else	
		return KIcon("run");
}

void Hebe::iplDevice( QTreeWidgetItem *pItem, int )
{
	if ( pItem->parent() > 0 )
		cmdInput( "IPL " + pItem->text( 0 ), true );
}

void Hebe::iplDevice( QListWidgetItem *pItem )
{
	cmdInput( "IPL " + pItem->text(), true );
}

void Hebe::cmdInput( QString cmd, bool internal )
{
	const int k = mpView->currentIndex();
	HebeVM *vm = mpImages->VM( k );
	if ( vm == 0 ) return;
	
	if ( vm->mStarted )
	{	
		cmd = cmd.trimmed();
		if ( cmd.startsWith( "IPL", Qt::CaseInsensitive ) )
			vm->mIPLed = true;
		
		if ( (vm->mpCons->mpTabs->currentIndex() == 1) & !internal )
			cmd.prepend( "scp " );
		
		cmd += '\n';
		vm->mpProc->write( cmd.toAscii() );
	}
	
	vm->mpCons->mpConsInput->clear();
}

void Hebe::closePage()
{
	const int k = mpView->currentIndex();
	HebeVM *vm = mpImages->VM( k );
	if ( vm == 0 ) return;
	
	if ( !vm->mStarted )
	{	
		mpView->removePage( vm->mpPage );
		mpImages->removeVM( *vm );
		
		// kludge because of removePage() bug
		KPageView::FaceType oldFace = mpView->faceType();
		mpView->setFaceType( KPageView::Plain );
		mpView->setFaceType( oldFace );
	}
	else
		KMessageBox::sorry( this, i18n("The image is active, shut it down first!") );
}

void Hebe::changedTab( int )
{
/*	HebeVM *vm = mpImages->VM( k );
	
	if ( vm->mStarted )
	{	
	}*/
}

void Hebe::cbIndexChanged( int index )
{
	cmdInput( QString( " CPU %1" ).arg(index, 2, 16, QChar('0')), true );
}

void Hebe::pageChanged( KPageWidgetItem*, KPageWidgetItem* )
{
	const int k = mpView->currentIndex();
	HebeVM *vm = mpImages->VM( k );
	if ( vm == 0 ) return;
	
	if ( vm->mStarted )
	{
		mpStart->setEnabled( false );
		mpStop->setEnabled( true );
		mpSuspend->setEnabled( true );
		mpIPL->setEnabled( true );
		mpStartAll->setEnabled( true );
		mpStopAll->setEnabled( true );
		mpRstInt->setEnabled( true );
		mpExtInt->setEnabled( true );
	}
	else
	{
		mpStart->setEnabled( true );
		mpStop->setEnabled( false );
		mpSuspend->setEnabled( false );
		mpIPL->setEnabled( false );
		mpStartAll->setEnabled( false );
		mpStopAll->setEnabled( false );
		mpRstInt->setEnabled( false );
		mpExtInt->setEnabled( false );
	}	

	loadSBcb( vm );
	setViewMode( vm );
}

void Hebe::dragEnterEvent( QDragEnterEvent *event )
{
	if ( event->mimeData()->hasUrls() )
		event->acceptProposedAction();
}

void Hebe::dropEvent( QDropEvent *event )
{
	KUrl url( event->mimeData()->text() );
	
	if ( url.scheme() == "file" )
	{	
		event->acceptProposedAction();
		addImage( url.toLocalFile() );
	}	
}

void Hebe::rmbItem( const QPoint &point )
{
	QAction *action;
	const int k = mpView->currentIndex();
	HebeVM *vm = mpImages->VM( k );
	if ( vm == 0 ) return;
	
	QTreeWidget *pTw;
	pTw = qobject_cast<QTreeWidget *>( QObject::sender() );
	
	if ( pTw->topLevelItemCount() == 0 ) return;
	
	action = mpPopup->exec( pTw->viewport()->mapToGlobal(point) );    
	
	if ( action == 0 )
		return;
	else if ( action->data().toString() == "selectall" )
		selectAllItems( pTw );
	else if ( action->data().toString() == "copy" )
		copySelection( pTw );
}

void Hebe::selectAllItems( QTreeWidget *pTw )
{
	pTw->selectAll();
}

void Hebe::copySelection( QTreeWidget *pTw )
{
	QList<QTreeWidgetItem *> selRows = pTw->selectedItems();
	QString selText;
	
	for ( int j = 0; j< selRows.count(); ++j )
		selText += selRows[j]->text( 0 ) + '\n';

	QClipboard *clipboard = QApplication::clipboard();
	clipboard->setText( selText );
}

void Hebe::hlTimeout()
{
	HebeVM *vm = mpImages->objMap( QObject::sender() );
	
	deviceDim( *vm );
}

void Hebe::setViewMode( HebeVM *vm )
{
	QSplitter *pSplt = vm->mpCons->mpSplitter;
	QList<int> sz = pSplt->sizes();
	
	if ( mViewMode == ICON )
	{
		vm->mpCons->mpDevIv->show();
		vm->mpCons->mpDevTw->hide();
		if ( sz[0] == 0 ) sz[0] = sz[1];
		if ( sz[0] == 0 ) sz[0] = 100;
	}	
	else
	{
		vm->mpCons->mpDevIv->hide();
		vm->mpCons->mpDevTw->show();
		if ( sz[1] == 0 ) sz[1] = sz[0];
		if ( sz[1] == 0 ) sz[1] = 100;
	}	
	
	pSplt->setSizes( sz );
}

void Hebe::viewIconSlot( bool )
{
	if ( mViewMode == ICON ) return;
		
	const int k = mpView->currentIndex();
	HebeVM *vm = mpImages->VM( k );
	if ( vm == 0 ) return;
	
	mViewMode = ICON;
	setViewMode( vm );
}

void Hebe::viewTreeSlot( bool )
{
	if ( mViewMode == TREE ) return;
		
	const int k = mpView->currentIndex();
	HebeVM *vm = mpImages->VM( k );
	if ( vm == 0 ) return;
	
	mViewMode = TREE;
	setViewMode( vm );
}

void Hebe::pageViewSlot( bool )
{
	if ( mpPageViewList == QObject::sender() )
		mpView->setFaceType( KPageView::Auto );
	else if ( mpPageViewTabbed == QObject::sender() )
		mpView->setFaceType( KPageView::Tabbed );
	else if ( mpPageViewTree == QObject::sender() )
		mpView->setFaceType( KPageView::Tree );
}

void Hebe::loadSBcb( HebeVM *vm )
{
	mpSBcb->clear();
	for ( int k = 0; k < vm->mNumCPUs; ++k )
		mpSBcb->addItem( QString( "CP%1" ).arg(k, 2, 16, QChar('0')) );
}

void Hebe::checkHerc()
{
  mProg = whereIs( "hercules" );
	if ( mProg.isEmpty() )
	{
		KMessageBox::detailedError( this, "Hercules program not found!",
			"Hercules is not installed. Please download from\n\n"
			"  http://www.hercules-390.org \n\n"
			"or install from your distro.");
		
		close();
	}		
}

void Hebe::checkResumeFile( HebeVM *vm )
{
	QString srfile( vm->mDir + "/hercules.srf.gz" );
	QFileInfo stat(srfile);
	
	if ( stat.exists() )
	{
		int rc = KMessageBox::questionYesNoCancel( vm->mpCons, 
								i18n("An image resume file is available. Choose an option:\n\n"
										 "   Resume - to attempt to resume a saved image\n"
										 "   Delete - to delete the resume file and continue\n"
										 "   Ignore - to ignore the file and continue"),
								i18n("Saved image available"), 
								KGuiItem(i18n("&Resume"), KIcon("edit-redo")), 
								KGuiItem(i18n("&Delete"), KIcon("edit-delete")), 
								KGuiItem(i18n("&Ignore"), KIcon("dialog-cancel")) );
		
		switch ( rc )
		{
			case KMessageBox::Yes:
				cmdInput( "resume", true );
				break;
				
			case KMessageBox::No:
				QFile::remove( srfile );
				break;
		}	
	}	
}

QString Hebe::whereIs( const QString &bin )
{
	char line[128]; 	  
	FILE *fp;  

	fp = popen( QString("whereis -b " + bin).toAscii(), "r" );
	QString tmp( fgets( line, sizeof( line ), fp ) );
  pclose(fp);        
	return tmp.section( ' ', 1, 1 ).trimmed();         
}

int Hebe::getNumCPUs( const QString &cnf )
{
	char line[128]; 	  
	FILE *fp;  

	fp = popen( QString("grep NUMCPU \"" + cnf + "\"").toAscii(), "r" );
	QString tmp( fgets( line, sizeof( line ), fp ) );
  pclose(fp);        
	
	int n;
	if ( tmp.isEmpty() )
		n = 1;
	else
		n = tmp.section( ' ', 1, 1, QString::SectionSkipEmpty ).toInt();         
	
	return n;
}

#include "hebe.moc"
