/*
 * hebe.h
 *
 * Copyright (C) 2008 %{AUTHOR} <%{EMAIL}>
 */
#ifndef HEBE_H
#define HEBE_H


#define MAXLINES 1000
#define HERC_CMDS "?,help,*,message,msg,msgnoh,hst,hao,log,logopt,version,quit,exit,cpu,start,stop,startall,stopall,cf,cfall,.reply,!message,ssd,ptt,i,ext,restart,archmode,loadparm,ipl,iplc,sysreset,sysclear,store,psw,gpr,fpr,fpc,cr,ar,pr,timerint,clocks,ipending,ds,r,v,u,devtmax,k,attach,detach,define,devinit,devlist,qd,automount,scsimount,cd,pwd,sh,cache,cckd,shrd,conkpalv,quiet,t,t+,t-,t?,s,s+,s-,s?,b,b+,b-,g,ostailor,pgmtrace,savecore,loadcore,loadtext,ldmod,rmmod,lsmod,lsdep,iodelay,ctc,toddrag,panrate,msghld,syncio,maxrates,defsym,script,cscript,evm,ecpsvm,aea,aia,tlb,sizeof,suspend,resume,herclogo,traceopt,cmdtgt,herc,scp,pscp,sf+dev,sf-dev,sfc,sfk,sfd,t,s,t,f"
#define TITLE "Hercules Image "
#define  ITB "imageToolBar"

#include <qtreewidget.h>
#include <qmenu.h>
#include <qbrush.h>
#include <qcombobox.h>
#include <qlabel.h>

#include <kaction.h>
#include <klineedit.h>
#include <klocale.h>
#include <kprocess.h>
#include <kxmlguiwindow.h>
#include <kconfig.h>
#include <kconfiggroup.h>
#include <ktabbar.h>

#include "hebevm.h"
#include "hebecons.h"
#include "hebesb.h"

class HebeView;
class HebeVM;
class HebeVmList;

/**
 * This class serves as the main window for Hebe.  It handles the
 * menus, toolbars, and status bars.
 *
 * @short Main window class
 * @author %{AUTHOR} <%{EMAIL}>
 * @version %{VERSION}
 */
class Hebe : public KXmlGuiWindow
{
		Q_OBJECT
		
public:
    /**
     * Default Constructor
     */
    Hebe();

    /**
     * Default Destructor
     */
    virtual ~Hebe();
		
		friend class HebeVmList;
		
		enum ViewMode{ ICON, TREE };
		
public slots:

protected:
		void dropEvent( QDropEvent* event );
		void dragEnterEvent( QDragEnterEvent* event );
		bool queryClose();
	
private slots:
    void fileAdd();
    void fileRemove();
		void readStdOut();
		void readStdErr();
		void changedTab( int );
		void startSlot();
		void iplSlot();
		void imageCmdSlot();
		void terminateSlot();
		void rmbItem( const QPoint& );
		void selectAllItems( QTreeWidget* tw );
		void copySelection( QTreeWidget* tw );
		void confOptions();
		void cmdInput( QString cmd, bool internal = false );
		void finishedProc( int, QProcess::ExitStatus );
		void iplDevice( QTreeWidgetItem*, int );
		void iplDevice( QListWidgetItem* );
		void pageChanged( KPageWidgetItem*, KPageWidgetItem* );
		void hlTimeout();
		void viewIconSlot( bool );
		void viewTreeSlot( bool );
		void pageViewSlot( bool );
		void cbIndexChanged( int );
	
private:
    void setupActions();
    void setupStatusBar();
    void addImage( const QString& );
		void saveOptions();
		void readOptions();
		QString whereIs( const QString& );
		void checkHerc();
		void startProc( HebeVM* );
    void closePage();
    void deviceAttach(HebeVM& vm, const QString& def);
    void deviceChange(HebeVM& vm, const QString& device);
    void deviceDim(HebeVM& vm);
    void deviceDetach(HebeVM& vm, const QString& device);
		KIcon getDevIcon( const QString& );
		void setViewMode( HebeVM* );
		void splitToolTip( QString& );
		void clearStatusBar();
		void shutDown();
		void checkResumeFile(HebeVM* vm);
		int getNumCPUs( const QString& );
		void loadSBcb( HebeVM* );
    void enableImageTB( const bool enable );
		
private:
    HebeView *mpView;
		HebeVmList *mpImages;
		HebeSB mSBar;
    KSharedConfigPtr mConfig;
		KConfigGroup *mpGenGroup;
		KAction *mpAdd;
		KAction *mpRemove;
		KAction *mpStart;
		KAction *mpSuspend;
		KAction *mpStop;
		KAction *mpTerminate;
		KAction *mpIPL;
		KAction *mpStartAll;
		KAction *mpStopAll;
		KAction *mpRstInt;
		KAction *mpExtInt;
		KAction *mpFont;
		KAction *mpEdit;
		KAction *mpEditIcon;
		KAction *mpViewIcon;
		KAction *mpViewTree;
		KAction *mpPageViewList;
		KAction *mpPageViewTree;
		KAction *mpPageViewTabbed;
		QActionGroup *mpViewGrp;
		QActionGroup *mpPageGrp;
		KLineEdit *mpLnEdit;
		QMenu *mpPopup;
		QString mProg;
		QBrush mRedB;
		QBrush mGreenB;
		QBrush mGrayB;
		QBrush mYellowB;
		QBrush mWhiteB;
		QBrush mLightBlueB;
		QBrush mMagentaB;
		QList<int> mPaneSize;
		QByteArray mColSizes;
		KCompletion *mpCompl;
		ViewMode mViewMode;
};

#endif // _HEBE_H_
